import logging

from datetime import datetime

import time

import typing

from strategy_bridge.bus import Record
from strategy_bridge.common import config
if typing.TYPE_CHECKING:
    from strategy_bridge.processors import BaseProcessor


logger = logging.getLogger(config.DEBUGGER_LOGGER_NAME)


def wrap_with_measurements(self: "BaseProcessor", func: typing.Callable[..., typing.Any], *args, **kwargs):
    before = time.time()
    result = None
    exception = None
    try:
        result = func(self, *args, **kwargs)
    except Exception as e:
        exception = e
    after = time.time()
    return before, after, result, exception


def debugger(func: typing.Callable[..., typing.Any]) -> typing.Callable:
    def wrapper(self: "BaseProcessor", *args, **kwargs) -> typing.Any:
        if self.should_debug:
            before, after, result, exception = wrap_with_measurements(self, func, *args, **kwargs)
            logger.info(f"[{self.__class__.__name__}] [Processing took {after - before:.2f} seconds] "
                        f"[Current timestamp: {datetime.fromtimestamp(after)}]")

            if exception:
                raise exception
            return result
        else:
            return func(self, *args, **kwargs)
    return wrapper


def record_debugger(func: typing.Callable[..., typing.Any]) -> typing.Callable:
    def wrapper(self: "BaseProcessor", record: Record, *args, **kwargs) -> typing.Any:
        assert isinstance(record, Record), "First parameter to decorated function should be of type Record"

        before, after, result, exception = wrap_with_measurements(self, func, record, *args, **kwargs)

        msg_start = f"[{self.__class__.__name__}]"
        msg_status = f"[{'Failure' if exception else 'Processed'} after {after - before:.2f} seconds]"
        msg_stats = f"[Delay from input record: {after-record.timestamp:.2f} seconds] " \
                    f"[Input record timestamp: {datetime.fromtimestamp(record.timestamp)}]"
        msg = f"{msg_start} {msg_status} {msg_stats}"
        logger.info(msg)

        if exception:
            raise exception
        return result
    return wrapper



import attr

from strategy_bridge.bus import DataReader, DataBus, DataWriter
from strategy_bridge.common import config
from strategy_bridge.pb.messages_robocup_ssl_wrapper_pb2 import SSL_WrapperPacket
from strategy_bridge.processors import BaseProcessor
from strategy_bridge.utils.debugger import debugger


@attr.s(auto_attribs=True)
class VisionCombiner(BaseProcessor):

    vision_reader: DataReader = attr.ib(init=False)
    combined_vision_writer: DataWriter = attr.ib(init=False)

    def initialize(self, data_bus: DataBus) -> None:
        super(VisionCombiner, self).initialize(data_bus)

        self.vision_reader = DataReader(self.data_bus, config.VISION_DETECTIONS_TOPIC)
        self.combined_vision_writer = DataWriter(self.data_bus, config.COMBINED_VISION_DETECTIONS_TOPIC, 30)
        self._ssl_converter = SSL_WrapperPacket()

    @debugger
    def process(self):
        for ssl_record in self.vision_reader.read_new():
            ssl_package = ssl_record.content
            ssl_package = self._ssl_converter.FromString(ssl_package)
            print(type(ssl_package))



import time

import logging
import typing
import attr

from abc import ABC, abstractmethod

from strategy_bridge.bus import DataBus


@attr.s(auto_attribs=True, kw_only=True)
class BaseProcessor(ABC):

    processing_pause: typing.Optional[float] = 1
    should_run_infinitely: bool = True
    reduce_pause_on_process_time: bool = False
    should_debug: bool = False
    logger: logging.Logger = logging.getLogger(__name__)
    data_bus: DataBus = attr.ib(init=False)
    is_initialized: bool = attr.ib(init=False, default=False)

    def initialize(self, data_bus: DataBus) -> None:
        self.data_bus = data_bus
        self.is_initialized = True

    def run(self) -> None:
        if not self.is_initialized:
            raise Exception("Processor should be initialized first")
        self.logger.info(f"Running processor: {self.__class__.__name__}")
        try:
            if self.should_run_infinitely:
                while True:
                    self.main_loop()
            else:
                self.process()
        except KeyboardInterrupt:
            self.logger.warning(f"Interrupted {self.__class__.__name__}. Finalizing processing")
            self.finalize()

    def main_loop(self):
        before = time.time()
        self.process()
        after = time.time()
        took = after - before
        if self.processing_pause:
            pause = self.processing_pause
            if self.reduce_pause_on_process_time:
                pause -= took
                if pause < 0:
                    self.logger.warning(
                        f"Processor {self.__class__.__name__} took {took:.2f} seconds "
                        f"with expected pause between runs for {self.processing_pause} seconds"
                    )
                    pause = 0
            time.sleep(pause)

    @abstractmethod
    def process(self) -> None:
        pass

    def finalize(self) -> None:
        pass



import struct
import attr
import numpy as np

from strategy_bridge.bus import DataReader, DataWriter, Record, DataBus
from strategy_bridge.common import config
from strategy_bridge.model.referee import RefereeCommand
from strategy_bridge.processors import BaseProcessor
from strategy_bridge.utils.debugger import record_debugger, debugger
from strategy_bridge.pb.messages_robocup_ssl_wrapper_pb2 import SSL_WrapperPacket


@attr.s(auto_attribs=True)
class PythonControllerTemplate(BaseProcessor):

    max_commands_to_persist: int = 20

    vision_reader: DataReader = attr.ib(init=False)
    referee_reader: DataReader = attr.ib(init=False)
    commands_writer: DataWriter = attr.ib(init=False)

    CAMERAS_COUNT: int = 4
    MAX_BALLS_IN_CAMERA: int = 64
    MAX_BALLS_IN_FIELD: int = CAMERAS_COUNT * MAX_BALLS_IN_CAMERA
    BALL_PACKET_SIZE: int = 3

    ROBOTS_MAX_COUNT: int = 32
    TEAM_ROBOTS_MAX_COUNT: int = ROBOTS_MAX_COUNT // 2
    SINGLE_ROBOT_PACKET_SIZE = 5
    ROBOT_TEAM_PACKET_SIZE: int = SINGLE_ROBOT_PACKET_SIZE * TEAM_ROBOTS_MAX_COUNT

    GEOMETRY_PACKET_SIZE: int = 2

    def initialize(self, data_bus: DataBus) -> None:
        super(PythonControllerTemplate, self).initialize(data_bus)
        self.vision_reader = DataReader(self.data_bus, config.VISION_DETECTIONS_TOPIC)
        self.box_feedback_reader = DataReader(self.data_bus, config.BOX_FEEDBACK_TOPIC)
        self.referee_reader = DataReader(self.data_bus, config.REFEREE_COMMANDS_TOPIC)
        self.commands_writer = DataWriter(self.data_bus, config.ROBOT_COMMANDS_TOPIC, self.max_commands_to_persist)
        self._ssl_converter = SSL_WrapperPacket()


    def get_last_referee_command(self) -> RefereeCommand:
        referee_commands = self.referee_reader.read_new()
        if referee_commands:
            return referee_commands[-1].content
        return RefereeCommand(0, 0, False)

    @debugger
    def process(self) -> None:
        balls = np.zeros(self.BALL_PACKET_SIZE * self.MAX_BALLS_IN_FIELD)
        robots_blue = np.zeros(self.ROBOT_TEAM_PACKET_SIZE)
        robots_yellow = np.zeros(self.ROBOT_TEAM_PACKET_SIZE)
        field_info = np.zeros(self.GEOMETRY_PACKET_SIZE)

        feedback = self.box_feedback_reader.read_new()
        if feedback:
            print(feedback[-1])

        for ssl_record in self.vision_reader.read_new():
            rules = self.process_ssl(ssl_record, field_info, balls, robots_blue, robots_yellow)
            b = bytes()
            rules = b.join((struct.pack('d', rule) for rule in rules))
            self.commands_writer.write(rules)

    @record_debugger
    def process_ssl(self, ssl_record: Record, field_info, balls, robots_blue, robots_yellow):
        ssl_package = ssl_record.content
        ssl_package = self._ssl_converter.FromString(ssl_package)
        geometry = ssl_package.geometry
        if geometry:
            field_info[0] = geometry.field.field_length
            field_info[1] = geometry.field.field_width

        detection = ssl_package.detection
        if not detection:
            return

        camera_id = detection.camera_id
        for ball_ind, ball in enumerate(detection.balls):
            balls[ball_ind + (camera_id - 1) * self.MAX_BALLS_IN_CAMERA] = camera_id
            balls[ball_ind + self.MAX_BALLS_IN_FIELD + (camera_id - 1) * self.MAX_BALLS_IN_CAMERA] = ball.x
            balls[ball_ind + 2 * self.MAX_BALLS_IN_FIELD + (camera_id - 1) * self.MAX_BALLS_IN_CAMERA] = ball.y

        # TODO: Barrier states
        for robot in detection.robots_blue:
            robots_blue[robot.robot_id] = camera_id
            robots_blue[robot.robot_id + self.TEAM_ROBOTS_MAX_COUNT] = robot.x
            robots_blue[robot.robot_id + self.TEAM_ROBOTS_MAX_COUNT * 2] = robot.y
            robots_blue[robot.robot_id + self.TEAM_ROBOTS_MAX_COUNT * 3] = robot.orientation

        for robot in detection.robots_yellow:
            robots_yellow[robot.robot_id] = camera_id
            robots_yellow[robot.robot_id + self.TEAM_ROBOTS_MAX_COUNT] = robot.x
            robots_yellow[robot.robot_id + self.TEAM_ROBOTS_MAX_COUNT * 2] = robot.y
            robots_yellow[robot.robot_id + self.TEAM_ROBOTS_MAX_COUNT * 3] = robot.orientation

        referee_command = self.get_last_referee_command()
        rules = [5.0] * 32 * 13
        return rules



import typing

import attr

from strategy_bridge.bus import Record, DataBus


@attr.s(auto_attribs=True)
class DataReader:

    data_bus: DataBus
    read_topic_name: str
    last_read_message_timestamp: float = 0

    def read_new(self) -> typing.List[Record]:
        records = self.data_bus.read_from_timestamp(
            self.read_topic_name, self.last_read_message_timestamp
        )
        if records:
            self.last_read_message_timestamp = records[-1].timestamp
        return records

    def read_last(self) -> typing.Optional[Record]:
        record = self.data_bus.read_top(self.read_topic_name, 1)
        if record:
            return record[0]
        return None

    def read_all(self) -> typing.List[Record]:
        return self.data_bus.read_all(self.read_topic_name)



import json

import attr
import websocket
from websocket import WebSocketApp

from strategy_bridge.bus import DataWriter, DataBus
from strategy_bridge.common import config
from strategy_bridge.processors import BaseProcessor
from strategy_bridge.utils.debugger import debugger


@attr.s(auto_attribs=True)
class BoxFeedbackCollector(BaseProcessor):

    processing_pause = 0.01
    max_records_to_persist: int = 30
    box_ip: str = "10.0.120.103"
    box_port: int = 8001
    box_route: str = "/api/webclient"
    reconnect_count: int = 5
    should_run_infinitely: bool = False
    records_writer: DataWriter = attr.ib(init=False)
    websocket: WebSocketApp = attr.ib(init=False)

    def initialize(self, data_bus: DataBus) -> None:
        super(BoxFeedbackCollector, self).initialize(data_bus)
        self.records_writer = DataWriter(self.data_bus, config.BOX_FEEDBACK_TOPIC, self.max_records_to_persist)

        self.websocket = websocket.WebSocketApp(
            f"ws://{self.box_ip}:{self.box_port}{self.box_route}",
            on_open=self.on_open,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close
        )

    def on_open(self, ws):
        pass

    def on_close(self, ws, close_status_code, close_msg):
        pass

    def on_message(self, ws, message):
        parsed_message = json.loads(message)
        self.records_writer.write(parsed_message)

    def on_error(self, ws, error):
        pass

    def finalize(self) -> None:
        self.websocket.close()

    @debugger
    def process(self):
        self.websocket.run_forever(reconnect=self.reconnect_count)



import attr
import zmq

from strategy_bridge.bus import DataReader, DataBus
from strategy_bridge.common import config
from strategy_bridge.processors import BaseProcessor
from strategy_bridge.utils.debugger import debugger


@attr.s(auto_attribs=True)
class RobotCommandsSender(BaseProcessor):

    commands_reader: DataReader = attr.ib(init=False)

    def initialize(self, data_bus: DataBus) -> None:
        super(RobotCommandsSender, self).initialize(data_bus)

        context = zmq.Context()
        self.socket = context.socket(zmq.PUB)
        self.socket.bind(f"tcp://*:{config.COMMANDS_PUBLISH_PORT}")

        self.commands_reader = DataReader(self.data_bus, config.ROBOT_COMMANDS_TOPIC)

    @debugger
    def process(self):
        commands = self.commands_reader.read_new()
        for command in commands:
            self.socket.send(command.content)



import logging
from multiprocessing import Process

from multiprocessing.managers import BaseManager

import typing

import attr

from strategy_bridge.bus import DataBus
from strategy_bridge.processors import BaseProcessor


class BridgeManager(BaseManager):
    pass


@attr.s(auto_attribs=True, kw_only=True)
class Runner:
    processors: typing.List[BaseProcessor]
    logger: logging.Logger = logging.getLogger(__name__)

    def run(self):
        BridgeManager.register('data_bus', DataBus)
        with BridgeManager() as manager:
            data_bus = manager.data_bus()
            processes = [
                Process(target=self.run_processor, args=(processor, data_bus)) for processor in self.processors
            ]
            for process in processes:
                process.start()
            try:
                for process in processes:
                    process.join()
            except KeyboardInterrupt:
                self.logger.warning("The application was interrupted")

    def run_processor(self, processor: BaseProcessor, data_bus: DataBus) -> None:
        processor.initialize(data_bus)
        processor.run()
